#include    <stdio.h>
#include    <stdlib.h>

#define MAX_N_STATES    99

static int verbose;

static const Instr no_instr = {"#-", ' ', none, none, none, 0};

typedef struct {
    const Instr *instr;
    int cost;
} Label;

typedef struct {
    Label label[max_location];
} State;
static State clean_state;

static State state_set[MAX_N_STATES];
static const int first_state = 1;
static int first_free_state = 1;
static int first_new_state = 1;

static int initial_state[N_BASIC_LOCATIONS];
static int next_state[N_OPERATORS][MAX_N_STATES][MAX_N_STATES];
static const Instr *rewrite_instr[N_ACTUAL_LOCATIONS][MAX_N_STATES];

static void
print_state(const State *s) {
    Location loc;

    printf("State: {");
    for (loc = none; loc < max_location; loc++) {
        const Label *lab = &s->label[loc];

        if (lab->instr) {
            printf(" %s->%s@%d ",
                lab->instr->name, location2string(loc), lab->cost
            );
        }
    }
    printf("}\n");
}

static void
print_tables(void) {
    int i;

    printf("State set:\n");
    for (i = 0; i < first_free_state; i++) {
        printf("S%02d = ", i);
        print_state(&state_set[i]);
    }
    printf("End of state set;\n\n");

    printf("Initial state table:\n");
    for (i = 0; i < N_BASIC_LOCATIONS; i++) {
        printf("%s: S%02d\n",
            location2string(basic_location_list[i]),
            initial_state[i]
        );
    }
    printf("End initial state table;\n\n");

    printf("Transition tables:\n");
    for (i = 0; i < N_OPERATORS; i++) {
        int s_vert, s_hor;

        printf(" %c  ", operator_list[i]);
        for (s_hor = first_state; s_hor < first_free_state; s_hor++) {
            printf("S%02d ", s_hor);
        }
        printf("\n");
        for (s_vert = first_state; s_vert < first_free_state; s_vert++) {
            printf("S%02d ", s_vert);
            for (s_hor = first_state; s_hor < first_free_state; s_hor++) {
                printf("S%02d ", next_state[i][s_vert][s_hor]);
            }
            printf("\n");
        }
        printf("\n");
    }
    printf("End of transition tables;\n\n");

    printf("Code generation tables:\n");
    printf("    ");
    for (i = first_state; i < first_free_state; i++) {
        printf("S%02d ", i);
    }
    printf("\n");
    for (i = 0; i < N_ACTUAL_LOCATIONS; i++) {
        int s;

        printf("%s ", location2string(actual_location_list[i]));
        for (s = first_state; s < first_new_state; s++) {
            const Instr *instr = rewrite_instr[i][s];

            printf("%s  ", instr ? instr->name : "  ");
        }
        printf("\n");
    }
    printf("End of code generation tables;\n\n");
}

static void
normalize_state(State *s) {
    Location loc;
    int least_cost = -1;

    for (loc = none; loc < max_location; loc++) {
        const Label *lab = &s->label[loc];

        if (lab->instr) {
            int cost = lab->cost;

            if (least_cost == -1 || cost < least_cost) {
                least_cost = cost;
            }
        }
    }

    for (loc = none; loc < max_location; loc++) {
        Label *lab = &s->label[loc];

        if (lab->instr) {
            lab->cost -= least_cost;
        }
    }
}

static int
states_are_equal(const State *s1, const State *s2) {
    Location loc;

    for (loc = none; loc < max_location; loc++) {
        const Label *lab_1 = &s1->label[loc];
        const Label *lab_2 = &s2->label[loc];

        if (lab_2->instr != lab_1->instr
        ||  lab_2->cost != lab_1->cost
        ) return 0;
    }
    return 1;
}

static int
state_pos(const State *s) {
    int pos;

    for (pos = 0; pos < first_free_state; pos++) {
        if (states_are_equal(s, &state_set[pos])) return pos;
    }

    if (first_free_state == MAX_N_STATES) {
        fprintf(stderr,
            "Out of states, realloc(state_set) not yet implemented\n"
        );
        exit(1);
    }

    state_set[first_free_state] = *s;
    return first_free_state++;
}

static void
suggest_instruction_to_state(State *state, const Instr *ip, int cost) {
    Label *lab = &state->label[ip->result];

    if (/* there was no corresponding label yet */
        lab->instr == 0
    ||  /* the old label was inferior */
        lab->cost > cost
    ) {
        lab->instr = ip;
        lab->cost = cost;
    }
}

static int
get_initial_state(Location loc) {
    State new_state = clean_state;
    const Instr *ip;

    /* first, the basic location can be left in place at no cost */
    new_state.label[loc].instr = &no_instr;
    new_state.label[loc].cost = 0;

    /* second, search for load instructions for this basic operand */
    for (ip = &instr_list[0]; ip->name != 0; ip++) {
        if (ip->operator == ' '
        && ip->operand == loc
        && ip->second_operand == none
        ) {
            suggest_instruction_to_state(&new_state, ip, ip->cost);
        }
    }

    normalize_state(&new_state);
    return state_pos(&new_state);
}

static void
init_state_set(void) {
    int i;

    for (i = 0; i < N_BASIC_LOCATIONS; i++) {
        const Location loc = basic_location_list[i];

        initial_state[i] = get_initial_state(loc);

        if (initial_state[i] == 0) {
            fprintf(stderr,
                "No load instruction for basic location %s\n",
                location2string(loc)
            );
        }
    }
}

static int
get_next_state(char op, int s1, int s2) {
    State new_state = clean_state;
    const Instr *ip;

    for (ip = &instr_list[0]; ip->name != 0; ip++) {
        const Label *lab_1 = &state_set[s1].label[ip->operand];
        const Label *lab_2 = &state_set[s2].label[ip->second_operand];

        if (ip->operator == op && lab_1->instr && lab_2->instr) {
            /* matching instruction found */
            suggest_instruction_to_state(
                &new_state, ip, lab_1->cost + lab_2->cost + ip->cost
            );
        }
    }

    normalize_state(&new_state);
    return state_pos(&new_state);
}

static void
two_operand_states(int s1, int s2) {
    int i;

    for (i = 0; i < N_OPERATORS; i++) {
        const char oper = operator_list[i];

        next_state[i][s1][s2] =
            get_next_state(oper, s1, s2);

        if (next_state[i][s1][s2] == 0) {
            fprintf(stderr,
                "No transition for node ('%c', S%02d, S%02d)\n",
                oper, s1, s2
            );
        }

        if (verbose) {
            printf("Cost conscious next state['%c', S%02d, S%02d] = S%02d\n",
                oper, s1, s2, next_state[i][s1][s2]
            );
            printf("\n");

            print_state(&state_set[s1]);
            printf("'%c'\n", oper);
            print_state(&state_set[s2]);
            printf("->\n");
            print_state(&state_set[next_state[i][s1][s2]]);
            printf("\n");
        }
    }
}

static void
expand_state_set(void) {
    while (first_new_state < first_free_state) {
        int s;

        for (s = first_state; s < first_new_state; s++) {
            two_operand_states(s, first_new_state);
        }
        for (s = first_state; s < first_new_state; s++) {
            two_operand_states(first_new_state, s);
        }
        two_operand_states(first_new_state, first_new_state);
        first_new_state++;
    }
}

static void
generate_rewrite_table(void) {
    int i;

    for (i = 0; i < N_ACTUAL_LOCATIONS; i++) {
        const Location loc = actual_location_list[i];
        int s;

        for (s = first_state; s < first_new_state; s++) {
            rewrite_instr[i][s] = state_set[s].label[loc].instr;
        }
    }
}

int
main(int argc, char *argv[]) {
    while (argc > 1 && argv[1][0] == '-') {
        switch (argv[1][1]) {
        case 'v':
            verbose = 1;
            break;
        default:
            break;
        }
        argc--, argv++;
    }

    init_state_set();
    expand_state_set();
    generate_rewrite_table();
    print_tables();

    return 0;
}
